package org.um.feri.spin_msc_generator.parser;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.um.feri.spin_msc_generator.message.MSCMessage;
import org.um.feri.spin_msc_generator.process.MSCProcess;

public class MSCParser {
	
	/**
	 * Class MSCParser is a part of the MSCGenerator class. It implements methods for reading one file
	 * into lines, which are afterwards processed by methods from other classes. With regular expressions
	 * MSCParser extracts lines, which are a point of interest for future generation of MSC diagram.
	 *
	 * @version 2.0
	 * @since MSCSpinGenerator 2.0 
	 * 
	 * @author UM - FERI, Laboratorij za mikroracunalniske sisteme,  Tim KOVSE
	 * 
	 * @see processLine
	 * 
	 * */
		
	/*LOGGER*/
	private Logger		logger = Logger.getLogger("MSCLogger");
	/*LOGGER*/
	
	/*REGEX*/
	/** tokens */
	private String[]	tokens	= null;
	private Pattern		pattern	= null;
	private Matcher		matcher	= null;
	/*REGEX*/
	
	/*PROCESS SINGLETON*/
	private MSCProcess	process 			= null;
	private String		processID			= null;
	private String		processName			= null;
	private String		processParentName	= null;
	private String		processParentID		= null;
	/*PROCESS SINGLETON*/
	
	/*MESSAGE SINGLETON*/
	private MSCMessage	message				= null;
	private String		messageMessage		= null;
	private String		messageValue		= null;
	private	Boolean		messageType			= null;
	private String		messageFrom			= null;
	private String		messageTo			= null;
	/*MESSAGE SINGLETON*/
	
	private String		line				= "";
	private int			lineID				= 0;
	
	private MSCMessage	tmpMessage			= null;
	private MSCMessage	tmpMessageFromList	= null;
	private MSCMessage	endMessage			= null;
	
	private String		messageProcessID	= null;
	private String[]	messageMessageValue = null;
	
	private boolean		startYsteps			= false;
	private boolean		prevIsNewProcess	= false;
	
	private ArrayList<MSCProcess>	processList		= new ArrayList();
	private ArrayList<MSCMessage> 	tmpMessageList	= new ArrayList();
	private ArrayList<MSCMessage>	messageList		= new ArrayList();
		
	/**
	 * 
	 * Constructor for new MSCParser
	 * 
	 * @param _fileName used for reading file, which must be parsed
	 * 
	 * */
	public MSCParser(String _fileName){
				
		BufferedReader bufferedReader = null;

		try{
			bufferedReader = new BufferedReader(new FileReader(_fileName));
			
			while ((line = bufferedReader.readLine()) != null){
				processLine(line);
			}

// List of Not Sent Messages.
//			MSCMessage msg;
//			for (int i=0; i<tmpMessageList.size(); i++){
//				msg = tmpMessageList.get(i);
//				System.out.println("Message name: " +msg.getMessage() + ", from:" + msg.getFrom() + " to:" + msg.getTo());
//			}
		}
		catch (IOException ex){
			logger.severe("Cannot read input file!");
			System.err.println(ex.getMessage());
			System.exit(1);	
		}
	}
	
	
	/**
	 * 
	 * Method processLine on the basis of regular expressions extracts informations used for creating processes
	 * and messages. The idea of this method is first to emit all unnecessary spaces and tabs (preprocessing line), 
	 * and than with regular expressions check if line meets requirements.
	 * 
	 * */
	
	private void processLine(String _line){
		_line = replaceMultiSpace(_line);
		
		tokens	= _line.split(" ");
		pattern = Pattern.compile("^(( )*[0-9]+:)$");
		matcher = pattern.matcher(tokens[0]);
				
		if (matcher.matches()){			
			if (tokens.length>=7){
				if (tokens[4].equals("creates")){
					
					if (startYsteps){
						lineID	= lineID+1;
						prevIsNewProcess = true;
					}
					
					processParentID 	= tokens[2];
					processParentName	= tokens[3];
					processID			= tokens[6];
					processName			= getName(tokens[7]);
					
					logger.info("Process: " + processName + "[" + processID + "] created.");
					
					process = new MSCProcess(processID, processName, processParentID, 
							processParentName, lineID);
					
					processList.add(process);
				}
			}
			if ((tokens.length>9)){
				if (tokens[9].equals("[values:")){

					startYsteps = true;
					
					if (prevIsNewProcess){					
						lineID	= lineID+3;
						prevIsNewProcess = false;
					}
					else {
						lineID	= lineID+1;
					}
					
					messageProcessID	= tokens[2];
					messageMessage 		= getMessage(tokens[10]);
					messageValue		= getMessageValue(tokens[10]);
					messageType			= getMessageType(tokens[10]);
					
					if (messageType){
						tmpMessage = new MSCMessage(messageMessage, messageValue, messageProcessID, lineID);
						tmpMessageList.add(tmpMessage);
						
						logger.info("Message " + messageMessage + " was sent from [" + processID + "].");
					}
					else {
						tmpMessageFromList = returnTmpListMessage(messageMessage, messageValue);
						endMessage = new MSCMessage(tmpMessageFromList.getMessage(),
								tmpMessageFromList.getValue(),
								tmpMessageFromList.getFrom(),
								messageProcessID,
								tmpMessageFromList.getSendID(),
								lineID);
						
						logger.info("Message " + messageMessage + " from " + tmpMessageFromList.getFrom() + " was received.");
						messageList.add(endMessage);
					}
				}
			}
		}
	}

	/**
	 * 
	 * Method replaceMultiSpace replaces multiple TABs and SPACEs that occur in line, which enables simpler 
	 * line processing in the next step.
	 * 
	 * */
	
	private String replaceMultiSpace(String _line){
		if (_line.startsWith(" ")){
			_line = _line.replaceFirst("[ ]+", "");
		}
		_line = _line.replaceAll("([ ]+)|([\t]+)", " ");
		return _line;
	}
	
	/**
	 * Name of a process is written between two brackets (<i>Example: (process)</i>), which are emitted by getName 
	 * method.
	 * 
	 * */
	
	private String getName(String _procName){
		_procName = _procName.substring(1, _procName.length()-1);
		return _procName;
	}
	
	/**
	 * Method getMessages accepts as a parameter a String which represents message and its value.
	 * When we are creating a new message we want to have message name and value separated. In Spin trail out
	 * file they are separated by one of two characters '!' or '?'. Result of this method is a message name. 
	 * 
	 * @return String
	 * 
	 * */
	
	private String getMessage(String _messageValue){
		_messageValue = _messageValue.substring(0, _messageValue.length()-1);
		
		if (_messageValue.contains("!")){
			messageMessageValue = _messageValue.split("!");
			_messageValue = messageMessageValue[0];
		}
		else {
			messageMessageValue = _messageValue.split("\\?");
			_messageValue = messageMessageValue[0];
		}
		return _messageValue;
	}
	
	/**
	 * Method getMessages accepts as a parameter a String which represents message and its value.
	 * When we are creating a new message we want to have message name and value separated. In Spin trail out
	 * file they are separated by one of two characters '!' or '?'. Result of this method is a message value. 
	 * 
	 * @return String
	 * 
	 * */
	
	private String getMessageValue(String _messageValue){
		_messageValue = _messageValue.substring(0, _messageValue.length()-1);
		
		if (_messageValue.contains("!")){
			messageMessageValue = _messageValue.split("!");
			_messageValue = messageMessageValue[1];
		}
		else {
			messageMessageValue = _messageValue.split("\\?");
			_messageValue = messageMessageValue[1];
		}
		return _messageValue;
	}
	
	/**
	 * Method getMessageType returns if message has been sent or recieved. If message was sent it returns true,
	 * otherwise it returns false.
	 * 
	 * @return boolean
	 * 
	 * */	
	private boolean getMessageType(String _messageValue){
		if (_messageValue.contains("!")){
			messageType = true;
		}
		else {
			messageType = false;
		}
		return messageType;
	}
	
	/**
	 * Method returnTmpListMessage returns temporary list of messages. Temporary list of messages is used
	 * for storing messages that have been sent but not yet received
	 * 
	 * @return temporary message list
	 * 
	 * */	
	
	private MSCMessage returnTmpListMessage(String _message, String _value){
		MSCMessage tmpListMessage = null;
		for (int i=0; i<tmpMessageList.size(); i++){
			tmpListMessage = tmpMessageList.get(i);
			
			if ((tmpListMessage.getMessage()).equals(_message)
					&& tmpListMessage.getValue().equals(_value)){
				tmpMessageList.remove(i);
				break;
			}
		}
		return tmpListMessage;
	}
	
	/**
	 * Method returns list of processes
	 * 
	 * */
	
	public ArrayList<MSCProcess> getProcessList(){
		return this.processList;
	}
	
	/**
	 * Method returns list of processes
	 * 
	 * */
	public ArrayList<MSCMessage> getMessageList(){
		return this.messageList;
	}
	
	public ArrayList<MSCMessage> getNotReceivedMessageList(){
		return this.tmpMessageList;
	}
}
